/*
 * Copyright (c) 2024-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibutils/release_containers.h>
#include "infiniband/ibutils/output_stream_primitives.h"
#include "ibdiagnet/ibdiagapp_types.h"

extern IBDiagClbck ibDiagClbck;

int IBDiag::BuildNVLAnycastLIDInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    const int BLOCK_CNT =
            (int)(sizeof(SMP_AnycastLIDInfo::block) /
                    sizeof(SMP_AnycastLIDInfo::block[0]));

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLAnycastLIDInfoGetClbck>;


    struct SMP_AnycastLIDInfo data = {};

    for (IBNode* p_hca : this->discovered_fabric.HCAs) {

        if (!p_hca) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_hca->getInSubFabric())
            continue;

        ib_extended_node_info* p_eni =
                this->fabric_extended_info.getSMPExtNodeInfo(p_hca->createIndex);

        if (!p_eni)
            continue;

        if (p_eni->node_type_extended != IB_GPU_NODE)
            continue;

        u_int32_t num_blocks = p_eni->AnycastLidCap / BLOCK_CNT + !!(p_eni->AnycastLidCap % BLOCK_CNT);

        for (u_int32_t i = 0; i < num_blocks; ++i) {

            progress_bar.push(p_hca);
            clbck_data.m_data1 = p_hca;
            clbck_data.m_data2 = (void*) (uintptr_t) i;

            direct_route_t * p_direct_route = this->GetDR(p_hca);
            if (!p_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s", p_hca->getName().c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            this->ibis_obj.SMPAnycastLIDInfoGetByDirect(p_direct_route, i /* block */, &data, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildNVLHBFConfig(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarPorts    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLHBFConfigGetClbck>;

    struct SMP_NVLHBFConfig data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLHBFConfigSupported))
            continue;

        // Port traversal
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            if (!p_port->is_data_worthy())
                continue;

            if (p_port->isSpecialPort())
                continue;

            progress_bar.push(p_port);
            clbck_data.m_data1 = p_port;

            direct_route_t * p_direct_route = this->GetDR(p_port);
            if (!p_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s", p_sw->getName().c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            this->ibis_obj.SMPNVLHBFConfigGetByDirect(p_direct_route, p_port->num, 0 /* profile_number_en */, &data, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildNVLContainAndDrainInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLContainAndDrainInfoGetClbck>;

    struct SMP_ContainAndDrainInfo data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsContainAndDrainSupported))
            continue;

        progress_bar.push(p_sw);
        clbck_data.m_data1 = p_sw;

        direct_route_t* p_direct_route = this->GetDR(p_sw);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_sw->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        for (int i = 0; i < (p_sw->numPorts + 127) / 128; ++i) {

            clbck_data.m_data2 = (void*)(uintptr_t)i;

            this->ibis_obj.SMPContainAndDrainInfoGetByDirect(p_direct_route,
                                                             (u_int8_t)i,
                                                             &data,
                                                             &clbck_data);
        }

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildNVLContainAndDrainPortState(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLContainAndDrainPortStateGetClbck>;

    struct SMP_ContainAndDrainPortState data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsContainAndDrainSupported))
            continue;

        progress_bar.push(p_sw);
        clbck_data.m_data1 = p_sw;

        direct_route_t* p_direct_route = this->GetDR(p_sw);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_sw->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        for (int i = 0; i < (p_sw->numPorts + 127) / 128; ++i) {

            clbck_data.m_data2 = (void*)(uintptr_t) i;

            this->ibis_obj.SMPContainAndDrainPortStateGetByDirect(p_direct_route,
                                                                  (u_int8_t)i,
                                                                  &data,
                                                                  &clbck_data);
        }

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildNVLClassPortInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLClassPortInfoGetClbck>;


    struct IB_ClassPortInfo data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        progress_bar.push(p_sw);
        clbck_data.m_data1 = p_sw;

        this->ibis_obj.NVLClassPortInfoGet(p_sw->getFirstLid(), 0 /* sl */, &data, &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}



int IBDiag::BuildNVLReductionInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLReductionInfoGetClbck>;

    struct NVLReductionInfo data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        if (!this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex))
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        progress_bar.push(p_sw);
        clbck_data.m_data1 = p_sw;

        this->ibis_obj.NVLReductionInfoGet(p_sw->getFirstLid(), 0 /* sl */, &data, &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildNVLReductionPortInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarPorts    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLReductionPortInfoGetClbck>;

    struct NVLReductionPortInfo data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        if (!this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex))
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        // Port traversal
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            if (!p_port->is_data_worthy())
                continue;

            if (p_port->isSpecialPort())
                continue;

            progress_bar.push(p_port);
            clbck_data.m_data1 = p_port;

            this->ibis_obj.NVLReductionPortInfoGet(
                    p_port->base_lid, 0 /* sl */, p_port->num, 0 /* profile */, &data, &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildNVLReductionForwardingTable(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarPorts    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLReductionForwardingTableGetClbck>;

    struct NVLReductionForwardingTable data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        if (!this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex))
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        // Check NVL ReductionInfo
        NVLReductionInfo * p_ri =
                this->fabric_extended_info.getNVLReductionInfo(p_sw->createIndex);

        if (!p_ri)
            continue;

        if (p_ri->reduction_fdb_top < MLID_START)
            continue;

        progress_bar.push(p_sw);
        clbck_data.m_data1 = p_sw;

        uint16_t total = (uint16_t)(p_ri->reduction_fdb_top - MLID_START + 1);

        // Block travel
        for (uint16_t n = 0; n < total; ++n) {

            clbck_data.m_data2 = (void*)(uintptr_t)n;

            this->ibis_obj.NVLReductionForwardingTableGet(p_sw->getFirstLid(),
                                                          0, // sl
                                                          n, // block
                                                          &data,
                                                          &clbck_data);
        } // blocks

        if (ibDiagClbck.GetState())
            goto exit;

    } // Switches

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildNVLPenaltyBoxConfig(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    const int MLID_CNT =
            (int)(sizeof(NVLPenaltyBoxConfig::mlid) /
                    sizeof(NVLPenaltyBoxConfig::mlid[0]));

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLPenaltyBoxConfigClbck>;

    struct NVLPenaltyBoxConfig data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        //
        // Check ClassPort Info
        //
        IB_ClassPortInfo * p_cpi =
                this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex);

        if (!p_cpi)
            continue;

        if (!(p_cpi->CapMsk & (1 << NVL_CPI_CapMas_IsPenaltyBoxSupported)))
            continue;

        // Check NVL ReductionInfo
        NVLReductionInfo * p_ri =
                this->fabric_extended_info.getNVLReductionInfo(p_sw->createIndex);

        if (!p_ri)
            continue;

        if (p_ri->reduction_fdb_top < MLID_START)
            continue;

        int total    = p_ri->reduction_fdb_top - MLID_START + 1;
        int blocks   = total / MLID_CNT + !!(total % MLID_CNT);

        for (int i = 0; i < blocks; ++i) {

            progress_bar.push(p_sw);

            clbck_data.m_data1 = p_sw;
            clbck_data.m_data2 = (void*) (uintptr_t) i;

            this->ibis_obj.NVLPenaltyBoxConfigGet(
                    p_sw->getFirstLid(), 0 /* sl */, (uint8_t)i /* block */, &data, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildNVLReductionConfigureMLIDMonitors(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarPorts    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLReductionConfigureMLIDMonitorsGetClbck>;

    struct NVLReductionConfigureMLIDMonitors data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        if (!this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex))
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        // Port traversal
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            if (!p_port->is_data_worthy())
                continue;

            if (p_port->isSpecialPort())
                continue;

            progress_bar.push(p_port);
            clbck_data.m_data1 = p_port;

            this->ibis_obj.NVLReductionConfigureMLIDMonitorsGet(
                    p_port->base_lid, 0 /* sl */, p_port->num, 0 /* profile */, &data, &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildNVLReductionCounters(list_p_fabric_general_err& errors, bool is_reset)
{
    IBDIAG_ENTER;

    const int MLID_CNT =
            (int)(sizeof(NVLReductionConfigureMLIDMonitors::mlid) /
                    sizeof(NVLReductionConfigureMLIDMonitors::mlid[0]));

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarPorts    progress_bar;
    clbck_data_t        clbck_data;

    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLReductionCountersGetClbck>;

    struct NVLReductionCounters data = {};

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_sw->getInSubFabric())
            continue;

        if (!this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex))
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        // Port traversal
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            if (!p_port->is_data_worthy())
                continue;

            if (p_port->isSpecialPort())
                continue;

            struct NVLReductionConfigureMLIDMonitors *p_monitor =
                    this->fabric_extended_info.getNVLReductionConfigureMLIDMonitors(p_port->createIndex);

            if (!p_monitor)
                continue;

            // Temporary change: send only 0xFFFF to aggregate all MLIDs
            uint16_t temp_mlid[MLID_CNT] = {0xFFFF, 0, 0, 0, 0, 0, 0, 0};

            for (int n = 0; n < MLID_CNT; ++n) {

                // if (!p_monitor->mlid[n])
                if (!temp_mlid[n])
                    continue;

                data.port_select      = (u_int16_t)i;
                // data.mlid             = p_monitor->mlid[n];
                data.mlid             = temp_mlid[n];
                data.lean_pipe_select = 0x0F;

                progress_bar.push(p_port);

                clbck_data.m_data1 = p_port;
                clbck_data.m_data2 = (void*)(uintptr_t)n;

                if (is_reset) {
                    data.counter_select = 0xffffffff;
                    this->ibis_obj.NVLReductionCountersSet(p_port->base_lid,
                            0 /* sl */, &data, &clbck_data);
                } else
                    this->ibis_obj.NVLReductionCountersGet(p_port->base_lid,
                            0 /* sl */, &data, &clbck_data);


                if (ibDiagClbck.GetState())
                    goto exit;
            }
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildNVLReductionRoundingMode(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes    progress_bar;

    clbck_data_t        clbck_data;
    clbck_data.m_p_obj          = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::NVLReductionRoundingModeGetClbck>;

    struct NVLReductionRoundingMode data = {};

    for (IBNode* p_node : this->discovered_fabric.Switches) {

        if (!p_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_node->getInSubFabric())
            continue;

        if (!this->fabric_extended_info.getNVLClassPortInfo(p_node->createIndex))
            continue;

        // Is node support NVLink
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                        p_node, EnSMPCapIsNVLReductionManagementSupported))
            continue;

        progress_bar.push(p_node);
        clbck_data.m_data1 = p_node;

        this->ibis_obj.NVLReductionRoundingModeGet(p_node->getFirstLid(),
                                                   0, // sl
                                                   0, // port - currently reserved in the AM
                                                   0, // profile
                                                   &data,
                                                   &clbck_data);

        if (ibDiagClbck.GetState())
            goto exit;

    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();

        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpAnycastLIDInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    const int BLOCK_CNT =
            (int)(sizeof(SMP_AnycastLIDInfo::block) /
                    sizeof(SMP_AnycastLIDInfo::block[0]));

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_ANYCAST_LID_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "Block";

    for (int i = 0; i < BLOCK_CNT; ++i)
        sstream << ",anycast_lid" << i << ",properties" << i;

    sstream << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_hca : this->discovered_fabric.HCAs) {

        if (!p_hca)
            continue;

        ib_extended_node_info * p_eni =
                this->fabric_extended_info.getSMPExtNodeInfo(p_hca->createIndex);

        if (!p_eni)
            continue;

        if (p_eni->node_type_extended != IB_GPU_NODE)
            continue;

        int blocks = p_eni->AnycastLidCap / BLOCK_CNT + !!(p_eni->AnycastLidCap % BLOCK_CNT);

        for (int i = 0; i < blocks; ++i) {

            struct SMP_AnycastLIDInfo * p_data =
                    this->fabric_extended_info.getNVLAnycastLIDInfo(p_hca->createIndex, i);

            if (!p_data)
                continue;

            sstream.str("");
            sstream
                << PTR(p_hca->guid_get()) << ','
                << DEC(i);

            for (int n = 0; n < BLOCK_CNT; ++n)
                sstream << "," << DEC(p_data->block[n].anycast_lid) << "," << PTR(p_data->block[n].properties);

            sstream
                << endl;

            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_NVL_ANYCAST_LID_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLHBFConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_HBF_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum,"
               "fields_enable,"
               "hash_type,"
               "packet_hash_bitmask,"
               "seed"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches)
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            struct SMP_NVLHBFConfig *p_data =
                    this->fabric_extended_info.getNVLHBFConfig(p_port->createIndex);

            if (!p_data)
                continue;

            sstream.str("");
            sstream
                << PTR(p_sw->guid_get()) << ','
                << PTR(p_port->guid_get()) << ','
                << DEC(p_port->num) << ','
                << PTR(p_data->fields_enable) << ','
                << PTR(p_data->hash_type) << ','
                << PTR(p_data->packet_hash_bitmask) << ','
                << DEC(p_data->seed) << endl;

            csv_out.WriteBuf(sstream.str());
        }

    csv_out.DumpEnd(SECTION_NVL_HBF_CONFIG);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLContainAndDrainInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_CONTAIN_AND_DRAIN_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum,"
               "egress_port_state,"
               "ingress_port_state"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        // Is node support ContainAndDrain
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsContainAndDrainSupported))
            continue;

        int port_num = 1;

        for (int i = 0; i < (p_sw->numPorts + 127) / 128; ++i) {
            struct SMP_ContainAndDrainInfo *p_data =
                    this->fabric_extended_info.getContainAndDrainInfo(p_sw->createIndex, i);

            IBPort * p_port0 = p_sw->getPort(0);

            if (!p_data || !p_port0) {
                port_num += 128;
                continue;
            }

            uint64_t sw_guid    = p_sw->guid_get();
            uint64_t port_guid  = p_port0->guid_get();

            for (int j = 0; j < 64 && port_num <= p_sw->numPorts; ++j) {
                // unlike the spec, here each block is one byte in size,
                // and contains info about 2 ports.
                IBPort* p_port1 = p_sw->getPort((phys_port_t)port_num);
                if (p_port1 &&
                        p_port1->is_data_worthy() &&
                        !p_port1->isSpecialPort()) {
                    sstream.str("");
                    sstream
                        << PTR(sw_guid) << ','
                        << PTR(port_guid) << ','
                        << DEC(port_num) << ','
                        << DEC(p_data->blocks[j].egress_port_state1) << ','
                        << DEC(p_data->blocks[j].ingress_port_state1)
                        << endl;

                    csv_out.WriteBuf(sstream.str());
                }

                IBPort* p_port2 = p_sw->getPort((phys_port_t)(port_num + 1));
                if (p_port2 &&
                        p_port2->is_data_worthy() &&
                        !p_port2->isSpecialPort()) {
                    sstream.str("");
                    sstream
                        << PTR(sw_guid) << ','
                        << PTR(port_guid) << ','
                        << DEC(port_num + 1) << ','
                        << DEC(p_data->blocks[j].egress_port_state2) << ','
                        << DEC(p_data->blocks[j].ingress_port_state2)
                        << endl;

                    csv_out.WriteBuf(sstream.str());
                }

                port_num += 2;
            }
        }
    }

    csv_out.DumpEnd(SECTION_NVL_CONTAIN_AND_DRAIN_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLContainAndDrainPortStateToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_CONTAIN_AND_DRAIN_PORT_STATE))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum,"
               "egress_port_state,"
               "ingress_port_state"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        // Is node support ContainAndDrain
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_sw, EnSMPCapIsContainAndDrainSupported))
            continue;

        int port_num = 1;

        for (int i = 0; i < (p_sw->numPorts + 127) / 128; ++i) {
            struct SMP_ContainAndDrainPortState *p_data =
                    this->fabric_extended_info.getContainAndDrainPortState(p_sw->createIndex, i);

            IBPort* p_port0 = p_sw->getPort(0);

            if (!p_data || !p_port0) {
                port_num += 128;
                continue;
            }

            uint64_t sw_guid    = p_sw->guid_get();
            uint64_t port_guid  = p_port0->guid_get();

            for (int j = 0; j < 64 && port_num <= p_sw->numPorts; ++j) {
                // unlike the spec, here each block is one byte in size,
                // and contains info about 2 ports.
                IBPort* p_port1 = p_sw->getPort((phys_port_t)port_num);
                if (p_port1 &&
                        p_port1->is_data_worthy() &&
                        !p_port1->isSpecialPort()) {
                    sstream.str("");
                    sstream
                        << PTR(sw_guid) << ','
                        << PTR(port_guid) << ','
                        << DEC(port_num) << ','
                        << DEC(p_data->blocks[j].egress_port_state1) << ','
                        << DEC(p_data->blocks[j].ingress_port_state1)
                        << endl;

                    csv_out.WriteBuf(sstream.str());
                }

                IBPort* p_port2 = p_sw->getPort((phys_port_t)(port_num + 1));
                if (p_port2 &&
                        p_port2->is_data_worthy() &&
                        !p_port2->isSpecialPort()) {
                    sstream.str("");
                    sstream
                        << PTR(sw_guid) << ','
                        << PTR(port_guid) << ','
                        << DEC(port_num + 1) << ','
                        << DEC(p_data->blocks[j].egress_port_state2) << ','
                        << DEC(p_data->blocks[j].ingress_port_state2)
                        << endl;

                    csv_out.WriteBuf(sstream.str());
                }

                port_num += 2;
            }
        }
    }

    csv_out.DumpEnd(SECTION_NVL_CONTAIN_AND_DRAIN_PORT_STATE);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLClassPortInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_CLASS_PORT_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "BaseVersion,"
               "ClassVersion,"
               "CapMsk,"
               "CapMsk2,"
               "RespTimeValue,"
               "RedirectGID,"
               "RedirectTC,"
               "RedirectSL,"
               "RedirectFL,"
               "RedirectLID,"
               "RedirectPKey,"
               "RedirectQP,"
               "RedirectQKey,"
               "TrapGID,"
               "TrapTC,"
               "TrapSL,"
               "TrapFL,"
               "TrapLID,"
               "TrapPKey,"
               "TrapHL,"
               "TrapQP,"
               "TrapQKey"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw)
            continue;

        struct IB_ClassPortInfo *p_data =
                this->fabric_extended_info.getNVLClassPortInfo(p_sw->createIndex);

        if (!p_data)
            continue;

        sstream.str("");
        sstream
            << PTR(p_sw->guid_get()) << ','
            << DEC(p_data->BaseVersion) << ','
            << DEC(p_data->ClassVersion) << ','
            << PTR(p_data->CapMsk) << ','
            << PTR(p_data->CapMsk2) << ','
            << DEC(p_data->RespTimeValue) << ','
            << PTR(p_data->RedirectGID[0].High)
                << HEX(p_data->RedirectGID[0].Low)
                << HEX(p_data->RedirectGID[1].High)
                << HEX(p_data->RedirectGID[1].Low)
                << ','

            << DEC(p_data->RedirectTC) << ','
            << DEC(p_data->RedirectSL) << ','
            << DEC(p_data->RedirectFL) << ','
            << DEC(p_data->RedirectLID) << ','
            << PTR(p_data->RedirectPKey) << ','
            << PTR(p_data->RedirectQP) << ','
            << PTR(p_data->RedirectQKey) << ','
            << PTR(p_data->TrapGID[0])
                << HEX(p_data->TrapGID[1])
                << HEX(p_data->TrapGID[2])
                << HEX(p_data->TrapGID[3])
                << ','
            << DEC(p_data->TrapTC) << ','
            << DEC(p_data->TrapSL) << ','
            << DEC(p_data->TrapFL) << ','
            << DEC(p_data->TrapLID) << ','
            << PTR(p_data->TrapPKey) << ','
            << DEC(p_data->TrapHL) << ','
            << DEC(p_data->TrapQP) << ','
            << PTR(p_data->TrapQKey)
            << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_NVL_CLASS_PORT_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLReductionInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_REDUCTION_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "capability_mask,"
               "reduction_fdb_cap,"
               "hbf_group_cap,"
               "reduction_fdb_top,"
               "num_of_up_ports,"
               "num_of_down_ports,"
               "mcto_timer,"
               "penalty_box_change_trap_disable"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw)
            continue;

        struct NVLReductionInfo *p_data =
                this->fabric_extended_info.getNVLReductionInfo(p_sw->createIndex);

        if (!p_data)
            continue;

        sstream.str("");
        sstream
            << PTR(p_sw->guid_get())           << ','
            << PTR(p_data->capability_mask)    << ','
            << PTR(p_data->reduction_fdb_cap)  << ','
            << PTR(p_data->hbf_group_cap)      << ','
            << PTR(p_data->reduction_fdb_top)  << ','
            << DEC(p_data->num_of_up_ports)    << ','
            << DEC(p_data->num_of_down_ports)  << ','
            << DEC(p_data->mcto_timer)         << ','
            << DEC(p_data->penalty_box_change_trap_disable)
            << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_NVL_REDUCTION_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLReductionPortInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_REDUCTION_PORT_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum,"
               "capability_mask,"
               "port_direction_is_up,"
               "exclude_include_self"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches)
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            struct NVLReductionPortInfo *p_data =
                    this->fabric_extended_info.getNVLReductionPortInfo(p_port->createIndex);

            if (!p_data)
                continue;

            sstream.str("");
            sstream
                << PTR(p_sw->guid_get()) << ','
                << PTR(p_port->guid_get()) << ','
                << DEC(p_port->num) << ','
                << PTR(p_data->capability_mask) << ','
                << DEC(p_data->port_direction_is_up) << ','
                << DEC(p_data->exclude_include_self)
                << endl;

            csv_out.WriteBuf(sstream.str());
        }

    csv_out.DumpEnd(SECTION_NVL_REDUCTION_PORT_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLReductionForwardingTableToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    const int HBF_GPOUPS =
            (int)(sizeof(NVLReductionForwardingTable::hbf_group_id) /
                    sizeof(NVLReductionForwardingTable::hbf_group_id[0]));

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_REDUCTION_FORWARDING_TABLE))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "Block,"
               "mlid";

    for (int i = 0; i < HBF_GPOUPS; ++i)
        sstream << ",hbf_group_id" << i;

    sstream << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw)
            continue;

        SMP_SwitchInfo * p_si =
                this->fabric_extended_info.getSMPSwitchInfo(p_sw->createIndex);

        if (!p_si)
            continue;

        NVLReductionInfo * p_ri =
                this->fabric_extended_info.getNVLReductionInfo(p_sw->createIndex);

        if (!p_ri)
            continue;

        for (int n = 0; n < p_ri->reduction_fdb_top; ++n) {

            struct NVLReductionForwardingTable * p_data =
                            this->fabric_extended_info.getNVLReductionForwardingTable(p_sw->createIndex, n);

            if (!p_data)
                continue;

            sstream.str("");
            sstream
            << PTR(p_sw->guid_get()) << ','
            << DEC(n) << ','
            << PTR((lid_t)(0xc000 + n + p_si->MCastFDBCap));

            for (int x = 0; x < HBF_GPOUPS; ++x)
                sstream << "," << PTR(p_data->hbf_group_id[x]);

            sstream << endl;

            csv_out.WriteBuf(sstream.str());
        } // blocks
    } // nodes

    csv_out.DumpEnd(SECTION_NVL_REDUCTION_FORWARDING_TABLE);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLPenaltyBoxConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    const int MLID_CNT =
            (int)(sizeof(NVLPenaltyBoxConfig::mlid) /
                    sizeof(NVLPenaltyBoxConfig::mlid[0]));

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_PENALTY_BOX_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "Block";

    for (int i = 0; i < MLID_CNT; ++i)
        sstream << ",mlid" << i;

    sstream << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches) {

        if (!p_sw)
            continue;

        SMP_SwitchInfo * p_si =
                this->fabric_extended_info.getSMPSwitchInfo(p_sw->createIndex);

        if (!p_si)
            continue;

        NVLReductionInfo * p_ri =
                this->fabric_extended_info.getNVLReductionInfo(p_sw->createIndex);

        if (!p_ri)
            continue;

        int total    = p_si->MCastFDBCap + p_ri->reduction_fdb_top;
        int blocks   = total / MLID_CNT + !!(total % MLID_CNT);

        for (int i=0; i < blocks; ++i) {

            struct NVLPenaltyBoxConfig * p_data =
                    this->fabric_extended_info.getNVLPenaltyBoxConfig(p_sw->createIndex, i);

            if (!p_data)
                continue;

            sstream.str("");
            sstream
                << PTR(p_sw->guid_get()) << ','
                << DEC(i);

            for (int n = 0; n < MLID_CNT; ++n)
                sstream << "," << std::dec << +p_data->mlid[n];

            sstream
                << endl;

            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_NVL_PENALTY_BOX_CONFIG);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLReductionConfigureMLIDMonitorsToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    const int MLID_CNT =
        (int)(sizeof(NVLReductionConfigureMLIDMonitors::mlid) /
                    sizeof(NVLReductionConfigureMLIDMonitors::mlid[0]));

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_REDUCTION_CONFIGURE_MLID_MONITORS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum";

    for (int i = 0; i < MLID_CNT; ++i)
        sstream << ",mlid" << i;

    sstream << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches)
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            struct NVLReductionConfigureMLIDMonitors *p_data =
                    this->fabric_extended_info.getNVLReductionConfigureMLIDMonitors(p_port->createIndex);

            if (!p_data)
                continue;

            sstream.str("");
            sstream
                << PTR(p_sw->guid_get()) << ','
                << PTR(p_port->guid_get()) << ','
                << DEC(p_port->num);

            for (int n = 0; n < MLID_CNT; ++n)
                sstream << "," << PTR(p_data->mlid[n]);

            sstream
                << endl;

            csv_out.WriteBuf(sstream.str());
        }

    csv_out.DumpEnd(SECTION_NVL_REDUCTION_CONFIGURE_MLID_MONITORS);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLReductionCountersToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    const int MLID_CNT =
        (int)(sizeof(NVLReductionConfigureMLIDMonitors::mlid) /
                    sizeof(NVLReductionConfigureMLIDMonitors::mlid[0]));

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_REDUCTION_COUNTERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum,"
               "mlid_idx,"
               "lean_pipe_select,"
               "mlid,"
               "incoming_good_mc_read_req_pkt,"
               "incoming_good_mc_read_req_data,"
               "incoming_good_mc_write_req_pkt,"
               "incoming_good_mc_write_req_data,"
               "incoming_good_mc_read_resp_pkt,"
               "incoming_good_mc_read_resp_data,"
               "incoming_good_mc_write_resp_pkt,"
               "incoming_good_mc_write_resp_data,"
               "generated_mc_read_resp_pkt,"
               "generated_mc_read_resp_data,"
               "generated_mc_write_resp_pkt,"
               "generated_mc_write_resp_data,"
               "dropped_pkt,"
               "trapped_pkt,"
               "nvl_reduction_errors,"
               "nvl_reduction_penalty_box_errors,"
               "port_nvr_local_alm_id_allocations,"
               "port_nvr_global_alm_id_allocations";

    sstream << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_sw : this->discovered_fabric.Switches)
        for (u_int32_t i = 1; i <= p_sw->numPorts; ++i) {
            IBPort *p_port = p_sw->getPort((phys_port_t)i);

            if (!p_port)
                continue;

            for (int n = 0; n < MLID_CNT; ++n) {

                struct NVLReductionCounters *p_data =
                        this->fabric_extended_info.getNVLReductionCounters(p_port->createIndex, n);

                if (!p_data)
                    continue;

                sstream.str("");
                sstream
                    << PTR(p_sw->guid_get()) << ','
                    << PTR(p_port->guid_get()) << ','
                    << DEC(p_port->num) << ','
                    << DEC(n) << ','
                    << PTR(p_data->lean_pipe_select) << ','
                    << PTR(p_data->mlid) << ','
                    << DEC(p_data->incoming_good_mc_read_req_pkt) << ','
                    << DEC(p_data->incoming_good_mc_read_req_data) << ','
                    << DEC(p_data->incoming_good_mc_write_req_pkt) << ','
                    << DEC(p_data->incoming_good_mc_write_req_data) << ','
                    << DEC(p_data->incoming_good_mc_read_resp_pkt) << ','
                    << DEC(p_data->incoming_good_mc_read_resp_data) << ','
                    << DEC(p_data->incoming_good_mc_write_resp_pkt) << ','
                    << DEC(p_data->incoming_good_mc_write_resp_data) << ','
                    << DEC(p_data->generated_mc_read_resp_pkt) << ','
                    << DEC(p_data->generated_mc_read_resp_data) << ','
                    << DEC(p_data->generated_mc_write_resp_pkt) << ','
                    << DEC(p_data->generated_mc_write_resp_data) << ','
                    << DEC(p_data->dropped_pkt) << ','
                    << DEC(p_data->trapped_pkt) << ','
                    << DEC(p_data->nvl_reduction_errors) << ','
                    << DEC(p_data->nvl_reduction_penalty_box_errors) << ','
                    << DEC(p_data->port_nvr_local_alm_id_allocations) << ','
                    << DEC(p_data->port_nvr_global_alm_id_allocations)
                    << endl;

                csv_out.WriteBuf(sstream.str());
            }
        }

    csv_out.DumpEnd(SECTION_NVL_REDUCTION_COUNTERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpNVLReductionRoundingModeToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NVL_REDUCTION_ROUNDING_MODE))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "deterministic_mode,"
               "mixed_16_bit,"
               "native_16_bit,"
               "mixed_8_bit,"
               "native_8_bit"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (IBNode* p_node : this->discovered_fabric.Switches) {

        if (!p_node)
            continue;

        struct NVLReductionRoundingMode *p_data =
                this->fabric_extended_info.getNVLReductionRoundingMode(p_node->createIndex);

        if (!p_data)
            continue;

        sstream.str("");
        sstream << PTR(p_node->guid_get())          << ','
                << DEC(p_data->deterministic_mode)  << ','
                << DEC(p_data->mixed_16_bit)        << ','
                << DEC(p_data->native_16_bit)       << ','
                << DEC(p_data->mixed_8_bit)         << ','
                << DEC(p_data->native_8_bit)        << endl;

        csv_out.WriteBuf(sstream.str());

    } // nodes

    csv_out.DumpEnd(SECTION_NVL_REDUCTION_ROUNDING_MODE);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
