/*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */



#include <infiniband/ibdm/ARTraceRoute.h>

bool g_useSLVLPortGroup = true;
int g_clearedCache = 0;


ARTraceRoutePath ARTraceRouteInfo::sm_ARTraceRoutePath;
//ARTraceRouteNodesInfoMap ARTraceRouteNodeInfo::sm_ARTraceRouteNodesInfo;

// Trace a route from slid to dlid by LFT
int ARTraceRouteByLFT(IBFabric *p_fabric,
                      lid_t sLid , lid_t dLid,
                      ARTraceRouteInfo *&pArRouteInfo)
{
    IBPort *p_port = p_fabric->getPortByLid(sLid);
    IBNode *p_node;
    pArRouteInfo = NULL;

    // make sure:
    if ( ! p_port ) {
        cout << "-E- Provided source:" << sLid
        << " lid is not mapped to a port!" << endl;
        return(1);
    }

    sl_vl_t slvl;
    slvl.SL = p_port->p_node->getPSLForLid(dLid);
    if ( slvl.SL == IB_SLT_UNASSIGNED ) {
        cout << "-E- Failed to get SL for node:" << p_port->p_node->name
                 << " dlid:" <<  dLid << endl;
        return 1;
    }
    slvl.VL = 0;
    //for in port 0 use SL even if VL2VL is set (for SW)
    slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);

    if ( p_port->p_node->type != IB_SW_NODE ) {

        if (slvl.VL == IB_SLT_UNASSIGNED) {
            cout << "-E- Failed to get VL for node:" << p_port->p_node->name
                << " inPort:0 outPort:"<< (int)p_port->num
                << " SL:" << (int)slvl.SL <<   endl;
            return(1);
        }
        if (slvl.VL == IB_DROP_VL) {
            cout << "-E- Dead end at:" << p_port->p_node->name
                << " Drop VL inPort:0 outPort:"<< (int)p_port->num
                << " SL:" << (int)slvl.SL <<   endl;
            return 1;
        }
    }

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        cout << "--------------------------- TRACE PATH BY FDB -----------------------------" << endl;
        cout << "-V- Tracing from lid:" << sLid
            << " SL/VL:" << (int)slvl.SL << "/" << (int)slvl.VL
            << " to lid:" << dLid << endl;
    }

    // if the port is not a switch - go to the next switch:
    if ( p_port->p_node->type != IB_SW_NODE ) {
        // try the next one:
        if ( !p_port->p_remotePort ) {
            cout << "-E- Provided starting point is not connected !"
            << "lid:" << sLid << endl;
            return 1;
        }
        p_port = p_port->p_remotePort;
        p_node = p_port->p_node;
        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << "-V- Arrived at Node:" << p_node->name
            << " Port:" << (unsigned int)p_port->num
            << " SL/VL:" << (int)slvl.SL << "/" << (int)slvl.VL <<endl;
    } else {
        // it is a switch :
        p_node = p_port->p_node;
    }


    // verify we are finally of a switch:
    if ( p_node->type != IB_SW_NODE ) {
        cout << "-E- Provided starting point is not connected to a switch !"
        << "lid:" << sLid << endl;
        return 1;
    }

    ARTraceRouteInfo * p_currInfo =
        ARTraceRouteNodeInfo::getRouteInfo(p_port, slvl, dLid);

    if (p_currInfo == NULL)
        return -1;

    pArRouteInfo = p_currInfo;

    p_currInfo->setCurrentAttributes(p_port->num);
    ARTraceRouteInfo::pathPushFront(p_currInfo);
    int pathSize = 1;

    ARTraceRouteInfo *p_childInfo = NULL;
    // traverse:
    while ( !ARTraceRouteInfo::isPathEmpty() ) {
        p_currInfo = ARTraceRouteInfo::pathFront();

        if (p_childInfo != NULL) {
            //finish discovery of child in path update parent
            p_currInfo->updateRouteStatistics(p_childInfo);
            p_childInfo = NULL;
        }

        phys_port_t out_port = p_currInfo->getNextPort();

        // if route discovery was ended from this route info
        if ( out_port== IB_LFT_UNASSIGNED ) {
            p_childInfo = p_currInfo;
            p_currInfo->pathPopFront();
            pathSize--;

            p_currInfo->markReachedFLIDs();

            if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                p_currInfo->dumpRouteStatistics();
            //collect statistics, routes per hops
            continue;
        }

        // if the port number is 0 we must have reached the target
        // (and the target is a switch - currently unused)
        if ( out_port == 0 ) {
            uint16_t base_lid = 0;
            // get lid of any port of this node
            // use unsigned int index to prevent overwrap
            for ( unsigned int portNum = 0;
                (base_lid == 0) && (portNum <= p_node->numPorts); portNum++ ) {
                IBPort *p_port = p_node->getPort((phys_port_t)portNum);
                if ( p_port ) {
                    base_lid = p_port->base_lid;
                    break;
                }
            }
            if ( base_lid == 0 ) {
                cout << "-E- Fail to find node:" << p_node->name
                    << " base lid?" << endl;
                p_currInfo->addCount(AR_TRACE_ROUTE_DEAD_END);
                continue;
            }

            lid_t lidStep = (lid_t)(1 << p_port->lmc);
            if ( (base_lid > dLid) || (base_lid + lidStep - 1 < dLid) ) {
                cout << "-E- Dead end at port 0 of node:" << p_node->name << endl;
                p_currInfo->addCount(AR_TRACE_ROUTE_DEAD_END);
                continue;
            }
            p_currInfo->addGoodPath(0);
            continue;
        }

        // skip Asymmetric APorts
        IBNode* p_curr_node = p_currInfo->getARTraceRouteNodeInfo()->getNode();
        IBPort* p_out_port  = p_curr_node ? p_curr_node->getPort(out_port) : NULL;

        if (p_out_port &&
                p_out_port->p_aport &&
                !p_out_port->p_aport->isSymmetric()) {

            cout << "-I- Skip routing through Asymmetric APort at:" << p_curr_node->name
                 << " port:" << +out_port
                 << endl;

            continue;
        }

        // calc next node:
        ARTraceRouteInfo*  p_nextInfo =
            p_currInfo->getNextARTraceRouteInfo(out_port);
        if ( p_nextInfo == NULL )
            continue;
        ARTraceRouteInfo::pathPushFront(p_nextInfo);
        pathSize++;

        if ( pathSize > 256 ) {
            cout << "-E- Aborting ARTraceRouteByLFT after 256 hops ["
                 << "sLid:" << sLid << "; "
                 << (p_fabric->isFLID(dLid) ? "dFLID:" : "dLid:") << dLid
                 << "]." << endl;
            while ( !ARTraceRouteInfo::isPathEmpty() )
                ARTraceRouteInfo::pathPopFront();
            break;
        }
    }

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "---------------------------------------------------------------------------\n" << endl;

    if ( !ARTraceRouteInfo::isPathEmpty() ) {
        ARTraceRouteInfo::clearPath();
        return -1;
    }

    return (p_currInfo->isErrorInPath() ? -1 : 0);
}

void ARTraceRouteInfo::pathPushFront(ARTraceRouteInfo *p_arInfo) {
    sm_ARTraceRoutePath.push_front(p_arInfo);
    p_arInfo->m_pNodeInfo->visit();
}

void ARTraceRouteInfo::pathPopFront() {
    sm_ARTraceRoutePath.front()->m_currInPort = IB_LFT_UNASSIGNED;
    sm_ARTraceRoutePath.front()->m_currOutPort = IB_LFT_UNASSIGNED;
    sm_ARTraceRoutePath.front()->m_pNodeInfo->leave();
    sm_ARTraceRoutePath.pop_front();
}

phys_port_t ARTraceRouteInfo::getNextPort()
{
    if (m_portsListIter == m_portsList.end())
        return getSkippedPort();

    if (m_incIter) {
        ++m_portsListIter;
        if (m_portsListIter == m_portsList.end())
            return getSkippedPort();
    } else
        m_incIter = true;

    return *m_portsListIter;
}

/*
phys_port_t ARTraceRouteInfo::getCurrentPort()
{
    if (m_portsListIter == m_portsList.end())
        return IB_LFT_UNASSIGNED;

    return *m_portsListIter;
} 
*/ 


void ARTraceRouteInfo::updateRouteStatistics(
    ARTraceRouteInfo *p_childInfo)
{
    for (int i = 0 ; i < AR_TRACE_ROUTE_END; ++i)
        m_routeStatistics[i] += p_childInfo->m_routeStatistics[i];

    m_errorInPath |= p_childInfo->m_errorInPath;

    m_minHops = min(m_minHops, p_childInfo->m_minHops + 1);
    m_maxHops = max(m_maxHops, p_childInfo->m_maxHops + 1);

    m_reachedFLIDs.insert(p_childInfo->m_reachedFLIDs.begin(), p_childInfo->m_reachedFLIDs.end());
}

void ARTraceRouteInfo::addGoodPath(unsigned int hops, IBNode *p_nodeToTrack)
{
    m_routeStatistics[AR_TRACE_ROUTE_GOOD_PATH]++;
    m_minHops = min(m_minHops, hops);
    m_maxHops = max(m_maxHops, hops);

    if (p_nodeToTrack)
        m_reachedFLIDs.insert(p_nodeToTrack);
}

void ARTraceRouteInfo::markReachedFLIDs()
{
    for (set_pnode::iterator itr = m_reachedFLIDs.begin(); itr != m_reachedFLIDs.end(); itr++) {
         if (*itr)
             (*itr)->appData3.val = 1;
    }
}

void ARTraceRouteInfo::dumpRouteStatistics()
{
    cout << "-V- Found total of "
        << m_routeStatistics[AR_TRACE_ROUTE_GOOD_PATH]
        << " paths";

    if (m_routeStatistics[AR_TRACE_ROUTE_GOOD_PATH])
        cout << " with [" << m_minHops << ".." << m_maxHops << "] hops.";

    if (m_pNodeInfo)
        cout << " From Node:" << m_pNodeInfo->getNode()->name
            << " PLFT:" << (int)m_pLFT
            << " in Port:" << (int)m_inSLVLPortGroup
            << " to DLID:" << m_dLid << endl;
    else
        cout << endl;

    if (m_errorInPath) {
        cout << "-E- Found total of " ;

            if (m_routeStatistics[AR_TRACE_ROUTE_DEAD_END] &&
                m_routeStatistics[AR_TRACE_ROUTE_LOOP])

                cout << m_routeStatistics[AR_TRACE_ROUTE_DEAD_END]
                    << " dead end errors and "
                    << m_routeStatistics[AR_TRACE_ROUTE_LOOP]
                    << " loops errors." << endl;
            else if (m_routeStatistics[AR_TRACE_ROUTE_DEAD_END])
                cout << m_routeStatistics[AR_TRACE_ROUTE_DEAD_END]
                    << " dead end errors." << endl;
            else
                cout << m_routeStatistics[AR_TRACE_ROUTE_LOOP]
                    << " loops errors." << endl;
    }
}

ARTraceRouteInfo* ARTraceRouteInfo::findNextARTraceRouteInfo(
    phys_port_t out_port,lid_t dLid, bool &reachedDest) const
{

    reachedDest = false;

    IBNode *p_node = m_pNodeInfo->getNode();
    //HW prevents loopback on the same port
    if (g_useSLVLPortGroup)
    {
        phys_port_t port = p_node->getSLVLPortGroup(out_port);
        if (port == m_inSLVLPortGroup || port == IB_LFT_UNASSIGNED)
            return NULL;
    } else if (out_port == m_inSLVLPortGroup) {
        return NULL;
    }

    // get the port on the other side
    IBPort *p_port = p_node->getPort(out_port);
    if ( ! (p_port &&
            p_port->p_remotePort) ) {
        return NULL;
    }

    IBPort *p_remotePort = p_port->p_remotePort;
    IBNode *p_remoteNode = p_remotePort->p_node;

    // check if we are done:
    lid_t lidStep = (lid_t)(1 << p_remotePort->lmc);
    bool done = ((p_remotePort->base_lid <= dLid) &&
            (p_remotePort->base_lid+lidStep - 1 >= dLid));

    // check if the dlid is a vport lid
    if (!done) {
        if (p_remoteNode->type == IB_CA_NODE) {
            map_vportnum_vport::iterator vports_iter = p_remotePort->VPorts.begin();
            for ( ; vports_iter != p_remotePort->VPorts.end(); vports_iter++) {

                if (vports_iter->second->get_vlid() == dLid) {
                    // we found the destination vport lid.
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)

                        cout << "-V- Going out from Node: " << p_node->name
                             << " pLFT:" << (int)m_pLFT
                             << " on port:" << (unsigned int)out_port
                             << " Arrived at destination Node:" << p_remoteNode ->name
                             << " Port:" << +p_remotePort->num
                             << " vPort: " << vports_iter->second->getVPortNum()
                             << " Virtual DLID:" << m_dLid << endl;

                    done = true;
                    break;
                }
            }
        }
    }

    // check if the dlid is a flid of a router
    if (!done) {
        if (p_remoteNode->type == IB_RTR_NODE) {
            if (p_remoteNode->remote_enabled_flids.find(m_dLid) != p_remoteNode->remote_enabled_flids.end()) {

                // we found the destination flid.
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- Going out from Node: " << p_node->name
                         << " pLFT:" << (int)m_pLFT
                         << " on port:" << (unsigned int)out_port
                         << " Arrived at destination Node:" << p_remoteNode ->name
                         << " Port:" << +p_remotePort->num
                         << " FLID:" << m_dLid << endl;

                // mark this router as found
                p_remoteNode->appData3.val = 1;
                done = true;
            }
        }
    }

    if (done) {
        reachedDest = true;
        return NULL;
    }

    if (p_remoteNode->type != IB_SW_NODE) {
        if (p_remoteNode->type == IB_RTR_NODE && p_remoteNode->p_fabric->isFLID(dLid))
            cout << "-E- Dead end to flid:" << dLid
                 << " at router:" << p_remoteNode->name
                 << ". The flid is not enabled on the router" << endl;
        else
            cout << "-E- Invalid route to lid:" << p_remotePort->base_lid
                 << " instead of:" << dLid << endl;
        return NULL;
    }

    sl_vl_t oSLVL;
    p_node->getSLVL(m_inSLVLPortGroup, out_port, getInSLVL(), oSLVL);

    if (oSLVL.VL == IB_SLT_UNASSIGNED ||
        oSLVL.VL == IB_DROP_VL) {
        return NULL;
    }

    ARTraceRouteInfo * p_nextRouteInfo =
        ARTraceRouteNodeInfo::findRouteInfo(p_remotePort, oSLVL);

    return p_nextRouteInfo;
}

ARTraceRouteInfo* ARTraceRouteInfo::getNextARTraceRouteInfo(
    phys_port_t out_port)
{

    //HW prevents loopback on the same port
    if (m_currInPort==out_port) {
        m_skippedOutPort = m_currInPort;
        return NULL;
    }

    if (out_port == m_skippedOutPort)
        m_skippedOutPort = IB_LFT_UNASSIGNED;

    m_currOutPort = out_port;

    // get the port on the other side
    IBNode *p_node = m_pNodeInfo->getNode();
    IBPort *p_port = p_node->getPort(out_port);

    //if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
    //    cout << "-V- Calculate Going out on port:" << (unsigned int)out_port << endl;

    if ( ! (p_port &&
            p_port->p_remotePort &&
            p_port->p_remotePort->p_node) ) {
        cout << "-E- Dead end to "
            << (p_node->p_fabric->isFLID(m_dLid) ? "flid:" : "lid:") << m_dLid << " at:"
            << p_node->name << " through port:"
            << (unsigned int)out_port << endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }

    if (p_node->skipRoutingChecks)
      return NULL;

    IBPort *p_remotePort = p_port->p_remotePort;
    IBNode *p_remoteNode = p_remotePort->p_node;
    // check if we are done:
    lid_t lidStep = (lid_t)(1 << p_remotePort->lmc);
    bool done = ((p_remotePort->base_lid <= m_dLid) &&
            (p_remotePort->base_lid+lidStep - 1 >= m_dLid));

    if (done) {
        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << "-V- Going out from Node: " << p_node->name
                << " pLFT:" << (int)m_pLFT
                <<  " on port:" << (unsigned int)out_port
                << " Arrived at destination Node:" << p_remoteNode ->name
                << " DLID:" << m_dLid << endl;

    } else { // check if the dlid is a vlid on the phys port we reached.

        if (p_remotePort->p_node->type == IB_CA_NODE) {
            map_vportnum_vport::iterator vports_iter = p_remotePort->VPorts.begin();
            for ( ; vports_iter != p_remotePort->VPorts.end(); vports_iter++) {

                if (vports_iter->second->get_vlid() == m_dLid) {
                    // we found the destination vport lid.

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)

                        cout << "-V- Going out from Node: " << p_node->name
                             << " pLFT:" << (int)m_pLFT
                             <<  " on port:" << (unsigned int)out_port
                             << " Arrived at destination Node:" << p_remoteNode ->name
                             << " Port:" << +p_remotePort->num
                             << " vPort: " << vports_iter->second->getVPortNum()
                             << " Virtual DLID:" << m_dLid << endl;

                    done = true;
                    break;
                }
            }
        }
    }

    IBNode *p_router = NULL;
    // check if the dlid is a flid of a router
    if (!done) {
        if (p_remoteNode->type == IB_RTR_NODE) {
            if (p_remoteNode->remote_enabled_flids.find(m_dLid) != p_remoteNode->remote_enabled_flids.end()) {

                // we found the destination flid.
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- Going out from Node: " << p_node->name
                         << " pLFT:" << (int)m_pLFT
                         << " on port:" << (unsigned int)out_port
                         << " Arrived at destination Node:" << p_remoteNode ->name
                         << " Port:" << +p_remotePort->num
                         << " FLID:" << m_dLid << endl;

                // mark this router as found
                p_remoteNode->appData3.val = 1;
                done = true;
                p_router = p_remoteNode;
            }
        }
    }

    if (done) {
        addGoodPath(1, p_router);
        return NULL;
    }

    if (p_remoteNode->type != IB_SW_NODE) {
        if (p_remoteNode->type == IB_RTR_NODE && p_remoteNode->p_fabric->isFLID(m_dLid))
            cout << "-E- Dead end to flid:" << m_dLid
                 << " at router:" << p_remoteNode->name
                 << ". The flid is not enabled on the router" << endl;
        else
            cout << "-E- Invalid route to lid:" << p_remotePort->base_lid
                 << " instead of:" << m_dLid << endl;

        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }

    sl_vl_t oSLVL;
    p_node->getSLVL(m_currInPort, out_port, getInSLVL(), oSLVL);

    if (oSLVL.VL == IB_SLT_UNASSIGNED) {
        cout << "-E- Failed to get VL for node:" << p_port->p_node->name
            << " inPort:" <<  (int)m_currInPort
            << " outPort:"<< (int)out_port
            << " SL:" << (int)oSLVL.SL <<   endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }
    if (oSLVL.VL == IB_DROP_VL) {
        cout << "-E- Dead end at:" << p_port->p_node->name
            << " Drop VL inPort:" <<  (int)m_currInPort
            << " outPort:"<< (int)out_port
            << " SL:" << (int)oSLVL.SL <<   endl;

        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-V- Going out from Node: " << p_node->name
            << " pLFT:" << (int)m_pLFT
            << " on port:" << (unsigned int)out_port
            << " SL/VL:" << (int)oSLVL.SL << "/" << (int)oSLVL.VL
            << " Arrived at Node:" << p_remoteNode ->name
            << " DLID:" << m_dLid << endl;

    ARTraceRouteInfo * p_nextRouteInfo =
        ARTraceRouteNodeInfo::getRouteInfo(p_remotePort, oSLVL, m_dLid);

    if(isLoopInRoute(p_nextRouteInfo)) {
        addCount(AR_TRACE_ROUTE_LOOP);
        return NULL;
    }

    p_nextRouteInfo->setCurrentAttributes(p_remotePort->num);
    return p_nextRouteInfo;
}

bool ARTraceRouteInfo::isLoopInRoute(ARTraceRouteInfo *p_routeInfo)
{

    if (p_routeInfo->m_currInPort == IB_LFT_UNASSIGNED)
        return false;

    cout << "-E- Find loop in path on Node:" <<
        p_routeInfo->m_pNodeInfo->getNode()->name
        << " to DLID:" << p_routeInfo->getDLID() << endl;

    ARTraceRoutePath::reverse_iterator rit = sm_ARTraceRoutePath.rbegin();
    for(;rit != sm_ARTraceRoutePath.rend(); ++rit)
        if(*rit == p_routeInfo)
            break;

    if (rit == sm_ARTraceRoutePath.rend()) {
        cout << "-E- Failed to find info in path:" << endl;
        return true;
    }

    IBNode *p_node = NULL;
    IBPort *p_port = NULL;
    IBPort *p_remotePort = NULL;
    phys_port_t currentPort;

    for (;rit != sm_ARTraceRoutePath.rend(); ++rit ) {
        p_node = (*rit)->m_pNodeInfo->getNode();
        currentPort =  (*rit)->getCurrentPort();
        p_port = p_node->getPort(currentPort);
        if (p_port == NULL) {
            cout << "-E- Failed to get node: " << p_node->name 
                << " port: " << (int) currentPort << endl;
            return true;
        }
        p_remotePort = p_port->p_remotePort;

        cout << "  - From Node:" << p_node->name
                << " pLFT:" << (unsigned int) (*rit)->getPLFT()
                << " port:" << (unsigned int) currentPort
                << " to Node:" << p_remotePort->p_node->name
                << " port:" << (unsigned int) p_remotePort->num << endl;
    }

    return true;
}

//
// Change optimization for GCC 12.2
// We found that this function cannot be compiled by GCC 12.2.0 with optimization level -O2
// but compiled without problem with optimization levels -O0 -O1 -O3
// We temporary changed optimization level only for this function to the -O1 in RELEASE mode
//
#ifndef DEBUG
  #ifdef __GNUC__
    #if __GNUC_PREREQ(12,2)
      __attribute__ ((optimize(1)))
    #endif
  #endif
#endif

bool ARTraceRouteInfo::isDestinationLIDReachable(phys_port_t oldPort, phys_port_t newPort, lid_t dLid) const
{
    const IBPort *p_oldPort = m_pNodeInfo->getNode()->getPort(oldPort);
    const IBPort *p_newPort = m_pNodeInfo->getNode()->getPort(newPort);

    //the cache has to be cleaned if one of ports is IB_LFT_UNASSIGNED.
    if (!p_oldPort || !p_newPort)
        return false;

    if (!p_oldPort->p_remotePort || !p_newPort->p_remotePort || p_newPort->p_remotePort->p_node)
        return false;

    //check base lid for the new destination lid
    const IBPort *p_remotePort = p_newPort->p_remotePort;
    lid_t lidStep = (lid_t)(1 << p_remotePort->lmc);
    if (p_remotePort->base_lid <= dLid &&
            p_remotePort->base_lid+lidStep - 1 >= dLid)
        return true;

    //already checked for not beeing NULL
    const IBNode *p_remoteNode = p_remotePort->p_node;
    if (p_remoteNode->type == IB_CA_NODE) {
        for (map_vportnum_vport::const_iterator vports_iter = p_remotePort->VPorts.begin();
                vports_iter != p_remotePort->VPorts.end(); vports_iter++) {
            if (vports_iter->second->get_vlid() == dLid)
                // we found the destination vport lid.
                return true;
        }
    }

    if (p_remoteNode->type == IB_RTR_NODE &&
            p_remoteNode->remote_enabled_flids.find(dLid) != p_remoteNode->remote_enabled_flids.end())
            return true;

    return false;
}

bool ARTraceRouteInfo::convertDestLid(lid_t dLid)
{
    const IBFabric *p_fabric = m_pNodeInfo->getNode()->p_fabric;
    if (p_fabric->isFLID(dLid) && m_errorInPath)
        return false;

    if (m_useAR) {
        const IBNode *p_node = m_pNodeInfo->getNode();
        u_int16_t newLFTPortGroup = p_node->getARLFTPortGroupForLid(dLid,m_pLFT);
        //keep for debugging
        /*if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-D- to dLID:" << dLid << " convertDestLid cache old LFTPortGroup:" << m_arLFTPortGroup
                << " new LFTPortGroup:" << newLFTPortGroup
                << " node:" << p_node->name
                << endl;
        */

        bool isStaticBoth = p_node->isARPortGroupEmpty(m_arLFTPortGroup) &&
                            p_node->isARPortGroupEmpty(newLFTPortGroup);
        if (isStaticBoth) {
            phys_port_t oldPort = p_node->getLFTPortForLid(m_dLid, m_pLFT);
            phys_port_t newPort = p_node->getLFTPortForLid(dLid, m_pLFT);
            //even if both ports are IB_LFT_UNASSIGNED we can keep this cache.
            if (oldPort != newPort) {
                //checks lmc and lid, virtual port lid and FLID
                if (!isDestinationLIDReachable(oldPort, newPort, dLid)) {
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-D- cache convertDestLid:" << dLid
                            << " old port:" << +oldPort
                            << " new port:" << +newPort
                            << " on node:" << p_node->name
                            << endl;

                    return false;
                }
            }
        } else {
            if (newLFTPortGroup != m_arLFTPortGroup) {

                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-D- cache convertDestLid:" << dLid
                        << " old LFTPortGroup:" << m_arLFTPortGroup
                        << " new LFTPortGroup:" << newLFTPortGroup
                        << " on node:" << p_node->name
                        << endl;

                return false;
            }
        }
    } else if(m_outStaticPort != m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)) {
        //in case AR is disabled
        if (m_errorInPath) {
            /*
            cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                 << " new StaticPort:"
                 << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                 << " m_errorInPath:" << m_errorInPath
                 << endl;
                 */

            return false;
        }

        bool reachedDest = false;
        ARTraceRouteInfo *nextRouteInfo = findNextARTraceRouteInfo(
            m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT), dLid, reachedDest);

        if ( reachedDest ){
            if (m_minHops != 1){
                /*
                cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                     << " new StaticPort:"
                     << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                     << " m_minHops:" << m_minHops
                     << endl;
                     */
                return false;
            }
        } else if (nextRouteInfo) {
            if (((nextRouteInfo->m_minHops + 1) != m_minHops) ||
                ((nextRouteInfo->m_maxHops + 1) != m_maxHops)){
                /*
                cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                     << " new StaticPort:"
                     << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                     << " old m_minHops:" << m_minHops
                     << " new m_minHops:" << nextRouteInfo->m_minHops
                     << " old m_maxHops:" << m_maxHops
                     << " new m_maxHops:" << nextRouteInfo->m_maxHops
                     << endl;
                     */
                return false;
            }
            if ((getGoodPathCount() != nextRouteInfo->getGoodPathCount()) ||
                nextRouteInfo->m_errorInPath){
                /*
                cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                     << " new StaticPort:"
                     << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                     << " old GoodPath:" << getGoodPathCount()
                     << " new GoodPath:" << nextRouteInfo->getGoodPathCount()
                     << " new errorInPath:" << nextRouteInfo->m_errorInPath
                     << endl;
                     */
                return false;
            }
        } else {
            /*
            cout << "-D- convertDestLid findNextARTraceRouteInfo out port:"
                 << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                 << endl;
                 */
            return false;
        }
    }
/*
    cout << "-D- convertDestLid on routeInfo: " << this
         << " m_dLid: " << m_dLid
         << " resetting to new DLID:"   << dLid
         << endl;
*/
    m_dLid = dLid;
    return true;
}

void ARTraceRouteInfo::set(
    sl_vl_t inSLVL,
    u_int8_t inPort, u_int8_t inSLVLPortGroup,
    u_int8_t pLFT, lid_t dLid,
    ARTraceRouteNodeInfo *pNodeInfo)
{
    IBNode *p_node = pNodeInfo->getNode();
    bool useAR = false;
    if (p_node->isARActive(inSLVL) || p_node->isHBFActive(inSLVL))
        useAR = 1;

    //reset old data
    m_inSLVL = inSLVL;
    m_errorInPath = false;
    m_minHops = 0xFFFF;
    m_maxHops = 0;
    m_currInPort = IB_LFT_UNASSIGNED;
    m_currOutPort = IB_LFT_UNASSIGNED;
    m_skippedOutPort = IB_LFT_UNASSIGNED;
    m_incIter = false;
    clearRouteStatistics();
    m_reachedFLIDs.clear();

    m_inSLVLPortGroup = inSLVLPortGroup;
    m_pNodeInfo = pNodeInfo;
    m_pLFT = pLFT;
    m_dLid = dLid;
    m_useAR = useAR;

    m_arLFTPortGroup = IB_AR_LFT_UNASSIGNED;
    if(m_useAR)
        m_arLFTPortGroup =
            p_node->getARLFTPortGroupForLid(dLid,pLFT);

    m_outStaticPort = p_node->getLFTPortForLid(dLid, pLFT);

    p_node->getLFTPortListForLid(m_outStaticPort,
                                 m_arLFTPortGroup, m_portsList);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        char line[1024];
        p_node->getARGroupCfg(m_arLFTPortGroup, line);
        cout << "-V- set RouteInfo for Node: " << p_node->name
            << " SL/VL:" << (int)inSLVL.SL << "/" << (int)inSLVL.VL
            << " pLFT:" << (int)m_pLFT
            << " AR:" << (useAR ? "enabled" : "disabled")
            << " static out port:" << (int) m_outStaticPort
            << " group:" << m_arLFTPortGroup
            << " group members:"
            << line << endl;
    }

    m_portsListIter = m_portsList.begin();

    if (m_portsListIter == m_portsList.end()) {
        cout << "-E- Dead end to "
            << (p_node->p_fabric->isFLID(m_dLid) ? "flid:" : "lid:") << m_dLid << " at:"
            << p_node->name << " pLFT:" << (int)pLFT << endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
    }  else if (*m_portsListIter == inPort &&
                    m_portsList.size() == 1) {
        cout << "-E- Dead end (loopback) to "
            << (p_node->p_fabric->isFLID(m_dLid) ? "flid:" : "lid:") << m_dLid << " at:"
            << p_node->name << " pLFT:" << (int)pLFT << endl;
    }
}

ARTraceRouteInfo* ARTraceRouteNodeInfo::findInfo(
        IBPort *p_port, sl_vl_t inSLVL)
{
    IBNode *p_node = p_port->p_node;

    u_int8_t pLFT = p_node->getPLFTMapping(p_port->num, inSLVL);

    u_int8_t sl2vlPortGroup = p_port->num;
    if (g_useSLVLPortGroup)
        sl2vlPortGroup = p_node->getSLVLPortGroup(p_port->num);

    if (m_routeInfoCollection[inSLVL.VL].size() <= inSLVL.SL ||
            m_routeInfoCollection[inSLVL.VL][inSLVL.SL].size() <= sl2vlPortGroup ||
            m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup].size() <= pLFT)
        return NULL;

    return &m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup][pLFT];
}

ARTraceRouteInfo* ARTraceRouteNodeInfo::getInfo(
        IBPort *p_port, sl_vl_t inSLVL, lid_t dLid)
{
    IBNode *p_node = p_port->p_node;

    u_int8_t maxPLFT = p_node->getMaxPLFT();
    u_int8_t pLFT = p_node->getPLFTMapping(p_port->num, inSLVL);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-V- AR pLFT MAP"
            << " Port:" << (int)p_port->num
            << " and SL/VL:" << (int)inSLVL.SL << "/" << (int)inSLVL.VL
            << " to pLFT:" << (int)pLFT
            << " on Node: " << p_node->name << endl;

    u_int8_t maxSL2VLPortGroup = p_node->numPorts;
    u_int8_t sl2vlPortGroup = p_port->num;
    if (g_useSLVLPortGroup)
        sl2vlPortGroup = p_node->getSLVLPortGroup(p_port->num);

    if (m_routeInfoCollection[inSLVL.VL].size() <= IBNode::getMaxSL())
        m_routeInfoCollection[inSLVL.VL].resize(IBNode::getMaxSL() + 1);

    if (m_routeInfoCollection[inSLVL.VL][inSLVL.SL].size() <= maxSL2VLPortGroup)
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL].resize(maxSL2VLPortGroup + 1);

    if (m_routeInfoCollection[inSLVL.VL][inSLVL.SL].size() <= sl2vlPortGroup)
        return NULL;

    if (m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup].size() <= maxPLFT)
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup].resize(maxPLFT + 1);

    ARTraceRouteInfo &info = m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup][pLFT];

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-D-"
            << " sl2vlPortGroup:" << (int)sl2vlPortGroup
            << " pLFT:" << (int)pLFT
            << " returns:" << (info.isSet() ? "old" : "new")
            << " routeInfo:" << &info << endl;

    if (!info.isSet()) { //if new info
        info.set(inSLVL, p_port->num, sl2vlPortGroup, pLFT, dLid, this);
        m_usedRouteInfo.push_back(&info);
    }

    return &info;
}


//allocate neaded memory
int ARTraceRouteNodeInfo::prepare(IBFabric *p_fabric)
{
  // go over all nodes in the fabric
  for (set_pnode::iterator nI = p_fabric->Switches.begin();
      nI != p_fabric->Switches.end(); nI++ ) {

    IBNode *p_node = *nI;

    // allocate ARTraceRouteNodeInfo
    p_node->appData1.ptr = new ARTraceRouteNodeInfo(p_node);
    if (p_node->appData1.ptr == NULL) {
        cout << "-E- Failed to allocate memory for ARTraceRoute" << endl;
        return 1;
    }

    /* check switches for slid representative (only 1 CA for slids loop)
       conditions are:
         1. same in SL
         2. same in VL
         3. same slvl port group
         4. same plft
     */

    p_node->routing_cache.CA_slid = 0;
    p_node->routing_cache.representative = true;

    sl_vl_t slvl;
    uint8_t vl = 0;
    uint8_t sl2vlPortGroup = 0;
    uint8_t pLFT = 0;

    for ( phys_port_t pn = 1; pn <= p_node->numPorts; pn++ ) {
        IBPort *p_port = p_node->getPort((phys_port_t)pn);
        if ( !p_port || !p_port->p_remotePort ||
                !p_port->p_remotePort->p_node ||
                p_port->p_remotePort->p_node->type == IB_SW_NODE)
            continue;

        // skip ANs
        if (p_port->isSpecialPort() &&
                p_port->getSpecialPortType() != IB_SPECIAL_PORT_ROUTER)
            continue;

        // CA connected to this switch
        IBPort *ca_port = p_port->p_remotePort;

        p_node->routing_cache.CA_lids.push_back(ca_port->base_lid);

        if (IBNode::getUsePSL() == false &&     // same in SL
                p_node->routing_cache.representative == true) {
            slvl.SL = p_port->p_node->p_fabric->GetDefaultSL();
            slvl.VL = 0;

            if (p_node->routing_cache.CA_slid == 0) {    // first CA on switch
                p_node->routing_cache.CA_slid = ca_port->base_lid;

                vl = ca_port->p_node->getVL(0, ca_port->num, slvl);

                sl2vlPortGroup = p_port->num;
                if (g_useSLVLPortGroup)
                    sl2vlPortGroup = p_node->getSLVLPortGroup(p_port->num);

                pLFT = p_node->getPLFTMapping(p_port->num, slvl);

            } else {     // check if all values equal to all CAs

                uint8_t cur_vl = ca_port->p_node->getVL(0, ca_port->num, slvl);
                uint8_t cur_sl2vlPortGroup = p_port->num;
                if (g_useSLVLPortGroup)
                    cur_sl2vlPortGroup = p_node->getSLVLPortGroup(p_port->num);
                uint8_t cur_pLFT = p_node->getPLFTMapping(p_port->num, slvl);

                if (cur_vl != vl ||
                        cur_sl2vlPortGroup != sl2vlPortGroup||
                        cur_pLFT != pLFT) {
                    p_node->routing_cache.representative = false;
                }
            }
        } else
            p_node->routing_cache.representative = false;

    } // ports loop
  } // Switches loop

  return 0;
}

void ARTraceRouteNodeInfo::cleanup(IBFabric *p_fabric)
{
  // go over all nodes in the fabric
  for (set_pnode::iterator nI = p_fabric->Switches.begin();
      nI != p_fabric->Switches.end(); nI++ ) {

    IBNode *p_node = *nI;

    delete (ARTraceRouteNodeInfo*)p_node->appData1.ptr;
    p_node->appData1.ptr = NULL;
  }
}

void ARTraceRouteNodeInfo::clearDB(IBFabric *p_fabric)
{
  g_clearedCache++;
  if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-D- clearing DB cache" << endl;

  // go over all nodes in the fabric
  for (set_pnode::iterator nI = p_fabric->Switches.begin();
      nI != p_fabric->Switches.end(); nI++ ) {

    IBNode *p_node = *nI;

    ARTraceRouteNodeInfo *p_nodeInfo =
        (ARTraceRouteNodeInfo*)p_node->appData1.ptr;

    p_nodeInfo->visitCount = 0;

    UsedRouteInfo::iterator iter=p_nodeInfo->m_usedRouteInfo.begin();
    for ( ; iter != p_nodeInfo->m_usedRouteInfo.end() ; ++iter) {
        (*iter)->clear();
    }
    p_nodeInfo->m_usedRouteInfo.clear();
  }
}

void ARTraceRouteNodeInfo::checkDB(IBFabric *p_fabric, lid_t dlid)
{
  // go over all nodes in the fabric
  for (set_pnode::iterator nI = p_fabric->Switches.begin();
      nI != p_fabric->Switches.end(); nI++ ) {

    IBNode *p_node = *nI;

    ARTraceRouteNodeInfo *p_nodeInfo =
        (ARTraceRouteNodeInfo*)p_node->appData1.ptr;

    p_nodeInfo->visitCount = 0;

    UsedRouteInfo::iterator iter=p_nodeInfo->m_usedRouteInfo.begin();
    for ( ; iter != p_nodeInfo->m_usedRouteInfo.end() ; ++iter) {
        if(!(*iter)->convertDestLid(dlid)){
            clearDB(p_fabric);
            return;
        }
    }
  }
}
