/*
 * Copyright (c) 2023-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/ibis/ibis_common.h>
#include <infiniband/misc/regexp/regexp.h>
#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SimulateA15.h>

#define QUANTUM3_BM "SW_BLACK_MAMBA"
#define QUANTUM3_CR "SW_CROCODILE"
#define HCA_CX8     "HCA_CX8"

#define BM_EXPECTED_PORTS   149
#define BM_MISSION_PORTS    144
#define BM_FIRST_FNM_PORT   145
#define BM_LAST_FNM_PORT    147
#define BM_EXTERNAL_PORT    148

#define CR_MISSION_PORTS    144
#define CR_FIRST_FNM_PORT   147
#define CR_LAST_FNM_PORT    148

int SimulateA15::SimulateSystemHeirarchyInfo(IBSystem* p_system) {
    if (!p_system || p_system->type.empty())
        return 0; 
    
    if (p_system->type == QUANTUM3_BM) {
        return SimulateBMHeirarchyInfo(p_system);
    }else if (p_system->type == QUANTUM3_CR) {
        return SimulateCRHeirarchyInfo(p_system);
    }else if (p_system->type == HCA_CX8) {
        return SimulateCX8HeirarchyInfo(p_system);
    } 

    return 0;
}

int SimulateA15::SimulateBMHeirarchyInfo(IBSystem* p_system) {
    for (auto& pair_name_node : p_system->NodeByName) {
        IBNode* p_node = pair_name_node.second;
        if (!p_node)
            continue;

        int asic_number = SimulateA15::GetAsicNumberFromNodeDescription(p_node);
        if (asic_number == -1)
            return -1;

        if (p_node->devId != DEV_ID_SW  || p_node->numPorts < BM_EXPECTED_PORTS)
            // not a switch ASIC node
            continue;

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort* p_port = p_node->getPort(i);
            if (!p_port || p_port->p_port_hierarchy_info || p_port->isSpecialPort())
                continue;

            PortHierarchyInfo* p_hier  = new PortHierarchyInfo(HIERARCHY_TEMPLATE_GUID_PORT_0x04);
            if (i <= BM_MISSION_PORTS) {
                p_hier->m_port_type        = 3;
                p_hier->m_cage             = (i + 1) / 2;
                p_hier->m_port             = (i + 1) % 2 + 1;
                p_hier->m_asic             = asic_number;
                p_hier->m_type             = 0;
                p_hier->m_plane            = asic_number;
                p_hier->m_num_of_planes    = 4;
                p_hier->m_aport            = i;
            } else if ( i >= BM_FIRST_FNM_PORT && i <= BM_LAST_FNM_PORT) {
                p_hier->m_port_type        = 2;
                p_hier->m_asic_name        = asic_number;
                p_hier->m_ibport           = i;
                p_hier->m_asic             = asic_number;
                p_hier->m_type             = 0;
            } else if (i == BM_EXTERNAL_PORT) {
                p_hier->m_port_type        = 1;
                p_hier->m_asic             = asic_number;
                p_hier->m_type             = 0;
                p_hier->m_plane            = asic_number;
                p_hier->m_num_of_planes    = 4;
                p_hier->m_aport            = i;
            }

            p_port->p_port_hierarchy_info = p_hier;
            p_hier->createLabel(p_node->type);
        }
    }

    return 0;
}

int SimulateA15::SimulateCRHeirarchyInfo(IBSystem* p_system) {
    for (auto& pair_name_node : p_system->NodeByName) {
        IBNode* p_node = pair_name_node.second;
        vector<ConnectionTypes> types(CR_MISSION_PORTS + 1);

        if (!p_node || p_node->devId != DEV_ID_SW ||
                p_node->numPorts < CR_MISSION_PORTS)
            continue;

        if (GetConnectionTypes(p_node, types))
            return -1;

        int asic_number = SimulateA15::GetAsicNumberFromNodeDescription(p_node);
        if (asic_number == -1)
            return -1;

        for (phys_port_t i = 1; i < p_node->numPorts; ++i) {
            IBPort* p_port = p_node->getPort(i);
            bool is_splitted = false;
            // if the port doesn't exist or hierarchy info already exists
            if (!p_port || p_port->p_port_hierarchy_info || p_port->isSpecialPort())
                continue;

            PortHierarchyInfo* p_hier  = new PortHierarchyInfo(HIERARCHY_TEMPLATE_GUID_PORT_0x04);

            if (i <= CR_MISSION_PORTS) {
                switch (types[i]) {
                    case PLNR_4X:
                        p_hier->m_num_of_planes = 4;
                        p_hier->m_plane = ((i - 1) % 4) + 1;
                        // APorts have fixed numbering (aport index for every 2 IBPorts)
                        p_hier->m_aport = ((i - 1) / 2) + 1;
                        // last 2 ports in a 4x APort need to have the same index as the first 2
                        if ((i - 1) % 4 >= 2)
                            p_hier->m_aport -= 1;
                        break;
                    case LEGACY_4X:
                        break;
                    case PLNR_2X:
                        p_hier->m_num_of_planes = 2;
                        p_hier->m_plane = ((i - 1) % 2) + 1;
                        // APorts have fixed numbering (aport index for every 2 IBPorts)
                        p_hier->m_aport = ((i - 1) / 2) + 1;
                        is_splitted = true;
                        break;
                    case LEGACY_2X:
                        is_splitted = true;
                        break;
                    default:
                        break;
                }

                p_hier->m_plane            = asic_number;
                p_hier->m_num_of_planes    = 4;
                p_hier->m_aport            = i;

                p_hier->m_port_type        = 3;
                p_hier->m_cage             = (i + 1) / 2;
                p_hier->m_port             = (i + 1) % 2 + 1;
                if (is_splitted)
                    p_hier->m_split        = (((i - 1) % 4) / 2) + 1;
                p_hier->m_asic             = asic_number;
                p_hier->m_type             = 0;
            } else if (i >= CR_FIRST_FNM_PORT && i <= CR_LAST_FNM_PORT) {
                // On ASIC 1, ports 147-148 are external FNM ports
                // On ASIC 2, ports 147-148 are internal FNM ports
                if (asic_number == 1) {
                    p_hier->m_port_type        = 1;
                    p_hier->m_asic             = asic_number;
                    p_hier->m_type             = 0;
                    // external FNM port may be planarized
                    if (IsPlanarizedPortList(p_node, CR_FIRST_FNM_PORT, CR_LAST_FNM_PORT)) {
                        p_hier->m_plane            = i - CR_FIRST_FNM_PORT + 1;
                        p_hier->m_num_of_planes    = 2;
                        p_hier->m_aport            = CR_MISSION_PORTS + 1;
                    }
                } else {
                    p_hier->m_port_type        = 2;
                    p_hier->m_asic_name        = asic_number + 255;
                    p_hier->m_ibport           = i;
                    p_hier->m_asic             = asic_number;
                    p_hier->m_type             = 0;
                }
            }

            p_port->p_port_hierarchy_info = p_hier;
            p_hier->createLabel(p_node->type);
        }

    }
    return 0;
}

int SimulateA15::SimulateCX8HeirarchyInfo(IBSystem* p_system) {
    for (auto& pair_name_node : p_system->NodeByName) {
        IBNode* p_node = pair_name_node.second;
        if (!p_node)
            continue;

        if (p_node->devId != DEV_ID_HCA)
            continue;

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort* p_port = p_node->getPort(i);
            if (!p_port || p_port->p_port_hierarchy_info)
                continue;

            PortHierarchyInfo* p_hier  = new PortHierarchyInfo(HIERARCHY_TEMPLATE_GUID_PORT_0x05);

            p_hier->m_bdf              = PortHierarchyInfo::bdf_t(1,2,3);
            p_hier->m_port_type        = 4;
            p_hier->m_cage             = 1;
            p_hier->m_port             = 1;
            p_hier->m_num_of_planes    = 4;
            p_hier->m_plane            = ((i - 1) % 4) + 1;
            p_hier->m_aport            = ((i - 1) / 4) + 1;

            p_port->p_port_hierarchy_info = p_hier;
            p_hier->createLabel(p_node->type);
        }
    }

    return 0;
}

int SimulateA15::GetAsicNumberFromNodeDescription(IBNode* p_node) {
    regExp namePattern("^([^/]*)/(\\w)([0-9]+)/");
    int rc = -1;
    
    
    rexMatch *p_match = nullptr;

    if (!p_node || p_node->name.empty())
        return rc;
    
    p_match = namePattern.apply(p_node->name.c_str());
    if (p_match && p_match->numFields() >= 3) {
        try {
            rc = std::stoi(p_match->field(3));
        } catch(...) {
            rc = -1;
        }
    }

    delete p_match;

    return rc;
}

int SimulateA15::GetConnectionTypes(IBNode* p_node, vector<ConnectionTypes>& connections) {
    if (!p_node || p_node->numPorts < CR_MISSION_PORTS)
        return -1;

    if (connections.size() < CR_MISSION_PORTS + 1) {
        connections.resize(CR_MISSION_PORTS + 1);
    }

    std::fill(connections.begin(), connections.end(), ConnectionTypes::INACTIVE);

    phys_port_t i = 1;
    while (i < CR_MISSION_PORTS) {
        IBPort* curr_port = p_node->getPort(i);
        if (!curr_port)
            return -1;

        if (!curr_port->p_remotePort) {
            i = (phys_port_t)(i + 1);
            continue;
        } else if ((i % 4) == 1 &&
                curr_port->get_internal_width() == IB_LINK_WIDTH_4X &&
                i < CR_MISSION_PORTS - 4) {
            // 4x legacy connection, fill the next 4 slots and advance i
            std::fill(connections.begin() + i, connections.begin() + i + 4,
                      ConnectionTypes::LEGACY_4X);
            i = (phys_port_t)(i + 4);
            continue;
        } else if ((i % 2) == 1 &&
                curr_port->get_internal_width() == IB_LINK_WIDTH_2X &&
                i < CR_MISSION_PORTS - 2) {
            // 2x legacy connection, fill the next 2 slots and advance i
            std::fill(connections.begin() + i, connections.begin() + i + 2,
                      ConnectionTypes::LEGACY_2X);
            i = (phys_port_t)(i + 2);
            continue;
        } else if ((i % 4) == 1 &&
                i < CR_MISSION_PORTS - 4 &&
                IsPlanarizedPortList(p_node, i, (phys_port_t)(i + 3))) {

            // 4x planarized connection, fill the next 4 slots and advance i
            std::fill(connections.begin() + i, connections.begin() + i + 4,
                      ConnectionTypes::PLNR_4X);
            i = (phys_port_t)(i + 4);
            continue;
        } else if ((i % 2) == 1 &&
                i < CR_MISSION_PORTS - 2 &&
                IsPlanarizedPortList(p_node, i, (phys_port_t)(i+1))) {
            // 2x planarized connection, fill the next 2 slots and advance i
            std::fill(connections.begin() + i, connections.begin() + i + 2,
                      ConnectionTypes::PLNR_2X);
            i = (phys_port_t)(i + 2);
            continue;
        }

        i = (phys_port_t)(i + 1);
    }

    return 0;
}

bool SimulateA15::IsPlanarizedPortList(IBNode* p_node, phys_port_t begin, phys_port_t end) {
    if (!p_node || !p_node->getPort(begin) ||
            !p_node->getPort(begin)->p_remotePort)
        return false;

    if (end <= begin || end >= p_node->numPorts)
        return false;

    uint64_t guid = p_node->getPort(begin)->p_remotePort->guid_get();

    for (phys_port_t i = (phys_port_t)(begin + 1); i <= end; ++i) {
        if (!p_node->getPort(i) ||
                !p_node->getPort(i)->p_remotePort ||
                p_node->getPort(i)->get_internal_width() != IB_LINK_WIDTH_1X ||
                p_node->getPort(i)->p_remotePort->guid_get() != guid)
            return false;
    }

    return true;
}
